@@ -4,6 +4,7 @@ import torch
 import torch.nn as nn
 from torch import optim
 import torch.nn.functional as F
+from okwugbe.okwugbe_asr import OkwugbeDatasetForCommonVoice
 from okwugbe.metrics import Metrics
 from okwugbe.decoder import Decoders
 from okwugbe.texttransform import TextTransform
@@ -12,6 +13,10 @@ from okwugbe.process import process
 from okwugbe.earlystopping import EarlyStopping
 import colorama
 import numpy as np
+from okwugbe.commonvoice import generate_character_set
+from IPython.display import Audio
+from IPython.core.display import display
+from livelossplot import PlotLosses
 
 # init the colorama module
 colorama.init()
@@ -48,7 +53,7 @@ def valid(model, device, test_loader, criterion, iter_meter, experiment, text_tr
 
     with torch.no_grad():
         for i, _data in enumerate(test_loader):
-            spectrograms, labels, input_lengths, label_lengths = _data
+            spectrograms, labels, input_lengths, label_lengths, _ = _data
             spectrograms, labels = spectrograms.to(device), labels.to(device)
 
             output = model(spectrograms)  # (batch, time, n_class)
@@ -74,18 +79,20 @@ def valid(model, device, test_loader, criterion, iter_meter, experiment, text_tr
                                                                                                        avg_cer,
                                                                                                        avg_wer)
     print(f"{GREEN}{valid_text}{RESET}")
-    return avg_wer
+    return avg_wer, experiment
 
 
 def train(model, device, train_loader, criterion, optimizer, scheduler, epoch, iter_meter, experiment, valid_loader,
-          best_wer, model_path, text_transform, early_stopping, batch_multiplier=1, grad_acc=False):
+          best_wer, model_path, text_transform, early_stopping, liveloss, batch_multiplier, grad_acc, display_plot):
     model.train()
+
     data_len = len(train_loader.dataset)
     train_loss = 0
     batch_train_loss = 0
-    print('Epoch {} ~ Training started'.format(epoch))
+    if not display_plot:
+        print('Epoch {} ~ Training started'.format(epoch))
     for batch_idx, _data in enumerate(train_loader):
-        spectrograms, labels, input_lengths, label_lengths = _data
+        spectrograms, labels, input_lengths, label_lengths, _ = _data
         spectrograms, labels = spectrograms.to(device), labels.to(device)
 
         optimizer.zero_grad()
@@ -93,13 +100,15 @@ def train(model, device, train_loader, criterion, optimizer, scheduler, epoch, i
         output = F.log_softmax(output, dim=2)
         output = output.transpose(0, 1)  # (time, batch, n_class)
         loss = criterion(output, labels, input_lengths, label_lengths)
+        loss.backward()
 
+        # torch.nn.utils.clip_grad_norm_(model.parameters(), clipping_value)
         # used if grad accumulation is used
         # This is the idea of grad accumulation to overcome memory issue
         if grad_acc:
             print('Using Gradient Accumulation')
             train_loss += loss.item() / (len(train_loader) * batch_multiplier)
-            if (batch_idx + 1) % batch_multiplier == 0:
+            if (batch_idx + 1) % batch_multiplier == 0 or batch_idx == data_len:
                 optimizer.step()
                 scheduler.step()
                 iter_meter.step()
@@ -110,23 +119,28 @@ def train(model, device, train_loader, criterion, optimizer, scheduler, epoch, i
         else:
             train_loss += loss.item() / len(train_loader)
 
-        loss.backward()
-        optimizer.step()
-        scheduler.step()
-        iter_meter.step()
+            optimizer.step()
+            scheduler.step()
+            iter_meter.step()
 
         if batch_idx % 5 == 0 or batch_idx == data_len:
-            text = 'Train Epoch {}: [{}/{} ({:.0f}%)] - Loss: {:.6f}'.format(epoch, batch_idx * len(spectrograms),
-                                                                             data_len,
-                                                                             100. * batch_idx / len(train_loader),
-                                                                             loss.item())
-            print(f"{YELLOW}{text}{RESET}")
+            if not display_plot:
+                text = 'Train Epoch {}: [{}/{} ({:.0f}%)] - Loss: {:.6f}'.format(epoch, batch_idx * len(spectrograms),
+                                                                                 data_len,
+                                                                                 100. * batch_idx / len(train_loader),
+                                                                                 loss.item())
+                print(f"{YELLOW}{text}{RESET}")
 
     experiment['loss'].append((train_loss, iter_meter.get()))
-    val_wer = valid(model, device, valid_loader, criterion, iter_meter, experiment, text_transform, epoch)  # wer
+    val_wer, experiment = valid(model, device, valid_loader, criterion, iter_meter, experiment, text_transform,
+                                epoch)  # wer
+    if display_plot:
+        logs = {'loss': experiment['loss'][-1][0], 'val_loss': experiment['val_loss'][-1][0],
+                'cer': experiment['cer'][-1][0], 'wer': experiment['wer'][-1][0]}
+        liveloss.update(logs)
+        liveloss.send()
 
-    for i in range(
-            1):  # Just to enable the 'break statement' - this will run once like a simple if/else statement
+    for i in range(1):  # Just to enable the 'break statement' - this will run once like a simple if/else statement
         early_stopping(val_wer, model, model_path)
         if early_stopping.early_stop:
             print("Early stopping")
@@ -141,9 +155,10 @@ def train(model, device, train_loader, criterion, optimizer, scheduler, epoch, i
             'best_wer': best_wer
         }, model_path)
     else:
-        print("No improvement in validation according to WER")
+        if not display_plot:
+            print("No improvement in validation according to WER")
 
-    return best_wer
+    return best_wer, experiment
 
 
 def test(model, device, test_loader, criterion, text_transform):
@@ -153,7 +168,7 @@ def test(model, device, test_loader, criterion, text_transform):
     test_cer, test_wer = [], []
     with torch.no_grad():
         for i, _data in enumerate(test_loader):
-            spectrograms, labels, input_lengths, label_lengths = _data
+            spectrograms, labels, input_lengths, label_lengths, audio_file_paths = _data
             spectrograms, labels = spectrograms.to(device), labels.to(device)
 
             output = model(spectrograms)  # (batch, time, n_class)
@@ -168,7 +183,8 @@ def test(model, device, test_loader, criterion, text_transform):
 
             for j in range(len(decoded_preds)):
                 print("Decoding Speech's Content")
-                print("Audio's Transcription: {}".format(decoded_preds[j]))
+                print(f'AUDIO PATH: {audio_file_paths[j]}')
+                display(Audio(str(audio_file_paths[j])))
                 test_cer.append(metrics.cer(decoded_targets[j], decoded_preds[j]))
                 test_wer.append(metrics.wer(decoded_targets[j], decoded_preds[j]))
                 current_prediction = "Decoded target: {}\nDecoded prediction: {}\n".format(decoded_targets[j],
@@ -183,15 +199,15 @@ def test(model, device, test_loader, criterion, text_transform):
 
 
 def main(model, train_path, test_path, validation_size, learning_rate, batch_size, epochs, experiment, cnn_layer,
-         rnn_layer,
-         model_path, rnn_dim, text_transform, batch_multiplier, grad_acc, n_class, n_feats, stride, dropout, optimizer,
-         patience):
+         rnn_layer, model_path, rnn_dim, text_transform, batch_multiplier, grad_acc, n_class, n_feats, stride, dropout,
+         optimizer,
+         patience, common_voice, freq_mask, time_mask, display_plot):
     if grad_acc:
         batch_size = batch_size // batch_multiplier
 
     print('Characters set: {}'.format(text_transform.char_map))
     print('Characters set length: {}'.format(len(text_transform.char_map)))
-    print('Number of classes: {}'.format(len(text_transform.char_map) + 1))
+    print('Number of classes: {}'.format(text_transform.get_num_classes()))
 
     hparams = {
         "n_cnn_layers": cnn_layer,
@@ -207,30 +223,46 @@ def main(model, train_path, test_path, validation_size, learning_rate, batch_siz
     }
 
     use_cuda = torch.cuda.is_available()
-    torch.manual_seed(7)
+
+    # torch.manual_seed(7) #this should be optional
     device = torch.device("cuda" if use_cuda else "cpu")
     print("Training with: {}".format(device))
 
-    train_dataset = OkwugbeDataset(train_path, test_path, "train", validation_size)
-    valid_dataset = OkwugbeDataset(train_path, test_path, "valid", validation_size)
-    test_dataset = OkwugbeDataset(train_path, test_path, "test", validation_size)
+    if common_voice['use_common_voice'] == True:
+        train_dataset = OkwugbeDatasetForCommonVoice(common_voice['lang'], "train", validation_size)
+        valid_dataset = OkwugbeDatasetForCommonVoice(common_voice['lang'], "valid", validation_size)
+        test_dataset = OkwugbeDatasetForCommonVoice(common_voice['lang'], "test", validation_size)
+
+    else:
+        train_dataset = OkwugbeDataset(train_path, test_path, "train", validation_size)
+        valid_dataset = OkwugbeDataset(train_path, test_path, "valid", validation_size)
+        test_dataset = OkwugbeDataset(train_path, test_path, "test", validation_size)
 
     kwargs = {'num_workers': 1, 'pin_memory': True} if use_cuda else {}
     train_loader = torch.utils.data.DataLoader(dataset=train_dataset,
                                                batch_size=hparams['batch_size'],
                                                shuffle=True,
-                                               collate_fn=lambda x: process.data_processing(x, text_transform, 'train'),
+                                               collate_fn=lambda x: process.data_processing(x, text_transform, 'train',
+                                                                                            n_feats=n_feats,
+                                                                                            freq_mask=freq_mask,
+                                                                                            time_mask=time_mask),
                                                **kwargs)
     valid_loader = torch.utils.data.DataLoader(dataset=valid_dataset,
                                                batch_size=hparams['batch_size'],
                                                shuffle=False,
-                                               collate_fn=lambda x: process.data_processing(x, text_transform, 'valid'),
+                                               collate_fn=lambda x: process.data_processing(x, text_transform, 'valid',
+                                                                                            n_feats=n_feats,
+                                                                                            freq_mask=freq_mask,
+                                                                                            time_mask=time_mask),
                                                **kwargs)
 
     test_loader = torch.utils.data.DataLoader(dataset=test_dataset,
                                               batch_size=hparams['batch_size'],
                                               shuffle=False,
-                                              collate_fn=lambda x: process.data_processing(x, text_transform, 'test'),
+                                              collate_fn=lambda x: process.data_processing(x, text_transform, 'test',
+                                                                                           n_feats=n_feats,
+                                                                                           freq_mask=freq_mask,
+                                                                                           time_mask=time_mask),
                                               **kwargs)
 
     model = model.to(device)
@@ -253,7 +285,7 @@ def main(model, train_path, test_path, validation_size, learning_rate, batch_siz
     if optimizer.lower() not in ['adamw', 'sgd', 'adam']:
         raise ValueError("Current optimizers supported: ['adamw', 'sgd', 'adam']")
 
-    criterion = nn.CTCLoss(blank=n_class - 1, zero_infinity=True).to(device)
+    criterion = nn.CTCLoss(blank=text_transform.get_blank_index(), zero_infinity=True).to(device)
 
     scheduler = optim.lr_scheduler.OneCycleLR(optimizer_, max_lr=hparams['learning_rate'],
                                               steps_per_epoch=int(len(train_loader)),
@@ -263,6 +295,8 @@ def main(model, train_path, test_path, validation_size, learning_rate, batch_siz
     iter_meter = IterMeter()
     best_wer = 1000
 
+    liveloss = PlotLosses()
+
     if os.path.exists(model_path):
         checkpoint = torch.load(model_path, map_location='cpu')
         model.load_state_dict(checkpoint['model_state_dict'])
@@ -272,28 +306,45 @@ def main(model, train_path, test_path, validation_size, learning_rate, batch_siz
         print("Training resumed from epoch {} with best WER == {}".format(epoch_saved + 1, best_wer))
         early_stopping = EarlyStopping(patience, model_path, best_wer, best_wer)
         for epoch in range(epoch_saved + 1, epochs + 1):
-            best_wer = train(model, device, train_loader, criterion, optimizer_, scheduler, epoch, iter_meter,
-                             experiment, valid_loader, best_wer, model_path, text_transform, early_stopping,
-                             batch_multiplier=batch_multiplier, grad_acc=grad_acc)
+            best_wer, experiment = train(model, device, train_loader, criterion, optimizer_, scheduler, epoch,
+                                         iter_meter,
+                                         experiment, valid_loader, best_wer, model_path, text_transform, early_stopping,
+                                         liveloss,
+                                         batch_multiplier, grad_acc, display_plot)
     else:
         early_stopping = EarlyStopping(patience, model_path, None, np.Inf)
         for epoch in range(1, epochs + 1):
-            best_wer = train(model, device, train_loader, criterion, optimizer_, scheduler, epoch, iter_meter,
-                             experiment, valid_loader, best_wer, model_path, text_transform, early_stopping,
-                             batch_multiplier=batch_multiplier, grad_acc=grad_acc)
+            best_wer, experiment = train(model, device, train_loader, criterion, optimizer_, scheduler, epoch,
+                                         iter_meter,
+                                         experiment, valid_loader, best_wer, model_path, text_transform, early_stopping,
+                                         liveloss, batch_multiplier, grad_acc, display_plot)
 
     print("Evaluating on Test data\n")
     test(model, device, test_loader, criterion, text_transform)
 
 
 class Train_Okwugbe:
-    def __init__(self, train_path, test_path, characters_set, n_cnn=5, n_rnn=3, rnn_dim=512, num_layers=1, n_feats=128,
+    def __init__(self, train_path=None, test_path=None, characters_set=None, n_cnn=5, n_rnn=3, rnn_dim=512,
+                 num_layers=1, n_feats=128,
                  in_channels=1, out_channels=32, kernel=3, stride=2, padding=1, dropout=0.1, with_attention=False,
-                 batch_multiplier=1, grad_acc=False, model_path='okwugbe_model', learning_rate=3e-5, batch_size=20,
-                 patience=20, epochs=500, optimizer='adamw', validation_size=0.2):
+                 batch_multiplier=1, grad_acc=False, model_path='okwugbe_model', learning_rate=3e-5, batch_size=80,
+                 patience=20, epochs=500, optimizer='adamw', validation_size=0.2, lang=None, use_common_voice=False,
+                 freq_mask=30, time_mask=100, display_plot=True):
+        if use_common_voice and lang is None:
+            raise Exception(
+                f'`lang` (language from Common Voice) must be specified if use_common_voice is set to True.')
+        self.common_voice = {'use_common_voice': use_common_voice, 'lang': lang}
+
+        if train_path is None and use_common_voice == False:
+            raise Exception(f'`train_path` cannot be None')
+        if test_path is None and use_common_voice == False:
+            raise Exception(f'`test_path` cannot be None')
         self.train_path = train_path
         self.test_path = test_path
-        self.characters_set = characters_set
+        if characters_set is None and use_common_voice == False:
+            raise Exception(
+                f'You need to specify a `characters set .txt` file or use Common Voice data by specifying `use_common_voice=True`.')
+        self.characters_set = characters_set if characters_set is not None else generate_character_set(lang)
         self.n_cnn = n_cnn
         self.n_rnn = n_rnn
         self.rnn_dim = rnn_dim
@@ -309,6 +360,7 @@ class Train_Okwugbe:
         self.batch_multiplier = batch_multiplier
         self.grad_acc = grad_acc
         self.model_path = model_path
+        self.model_path = self.model_path + '_' + lang if lang is not None else self.model_path
         self.learning_rate = learning_rate
         self.batch_size = batch_size
         self.patience = patience
@@ -316,9 +368,17 @@ class Train_Okwugbe:
         self.optimizer = optimizer
         self.validation_size = validation_size
         self.text_transform = TextTransform(self.characters_set)
-        self.n_class = len(self.text_transform.char_map) + 1
+        self.n_class = self.text_transform.get_num_classes()
         self.experiment = {'loss': [], 'val_loss': [], 'cer': [], 'wer': []}
 
+        # Speech augmentation
+        self.freq_mask = freq_mask
+        self.time_mask = time_mask
+
+        # Log the losses and metrics
+        self.logs = None
+        self.display_plot = display_plot
+
     def run(self):
         asr_model = SpeechRecognitionModel(self.n_cnn, self.n_rnn, self.rnn_dim, self.n_class, self.n_feats,
                                            self.in_channels, self.out_channels, self.kernel,
@@ -327,4 +387,5 @@ class Train_Okwugbe:
         main(asr_model, self.train_path, self.test_path, self.validation_size, self.learning_rate, self.batch_size,
              self.epochs, self.experiment, self.n_cnn, self.n_rnn, self.model_path, self.rnn_dim, self.text_transform,
              self.batch_multiplier, self.grad_acc, self.n_class, self.n_feats,
-             self.stride, self.dropout, self.optimizer, self.patience)
+             self.stride, self.dropout, self.optimizer, self.patience, self.common_voice, self.freq_mask,
+             self.time_mask, self.display_plot)
